#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>

#include "common.h"
#include "macros.h"
#include "buffer.h"
#include "io.h"
#include "timeout.h"
#include "xmalloc.h"
#include "hal-proto.h"
#include "hal-cmd-parser.h"
#include "hal-cmd.h"
#include "hal-peer.h"
#include "hal-runtime.h"

//#define DEBUG_HAL_RUNTIME

#define PEER_TIMEOUT_MSEC	3000	/* 3 seconds */

#ifdef DEBUG_HAL_RUNTIME
#define debug(fmt, arg...) fprintf(stderr, "debug: %s: "fmt CRLF, \
						__FUNCTION__, ##arg)
#else
#define debug(fmt, arg...) do {} while(0)
#endif

#define warn(fmt, arg...) fprintf(stderr, "warn: %s: "fmt CRLF, \
						__FUNCTION__,  ##arg)

#define info(fmt, arg...) fprintf(stderr, "info: %s: "fmt CRLF, \
						__FUNCTION__,  ##arg)

/* Structure represents a connection with the hal image manager. */
struct hal_runtime {
	struct hal_peer_base	peer;
	int			state;
	unsigned int		response_timeout;
	struct timeout		*tout;
	struct hal_runtime_callbacks callbacks;
};

static struct io_context	*io;
static struct hal_runtime	*hal_runtime;

int hal_runtime_state()
{
	return_val_if_fail(hal_runtime != NULL, -1);

	return hal_runtime->state;
}

void hal_runtime_disable()
{
	return_if_fail(hal_runtime != NULL);

	if (timeout_active(hal_runtime->tout))
		timeout_del(hal_runtime->tout);

	io_loop_stop(HAL_PEER_BASE(hal_runtime)->io);

	hal_runtime->state = HAL_RT_STATE_DISABLED;
}

static void hal_runtime_cmd(const char *cmd)
{
	struct hal_runtime *runtime;
	void (*request)(struct buffer *pkt);
	int new_state;

	runtime = hal_runtime;
	request = NULL;

	if (runtime->state == HAL_RT_STATE_DISABLED ||
	    runtime->state == HAL_RT_STATE_ERROR) {
		warn("%s request: runtime unexpected state", cmd);
		return;
	}

	if (cmd == HAL_CMD_CONNECT) {
		request = hal_cmd_connect_request;
		new_state = HAL_RT_STATE_CONN_WAIT;
	} else if (cmd == HAL_CMD_DISCONNECT) {
		request = hal_cmd_disconnect_request;
		new_state = HAL_RT_STATE_DISCONN_WAIT;
	} else if (cmd == HAL_CMD_CYCLE_BEGIN) {
		request = hal_cmd_cycle_begin_request;
		new_state = HAL_RT_STATE_CB_WAIT;
	} else if (cmd == HAL_CMD_CYCLE_RUN) {
		request = hal_cmd_cycle_run_request;
		new_state = HAL_RT_STATE_CR_WAIT;
	} else if (cmd == HAL_CMD_CYCLE_END) {
		request = hal_cmd_cycle_end_request;
		new_state = HAL_RT_STATE_CE_WAIT;
	}
	
	assert(request != NULL);

	request(HAL_PEER_BASE(runtime)->wb);
	runtime->state = new_state;

	timeout_add(runtime->tout, runtime->response_timeout);
	hal_peer_send_default(HAL_PEER_BASE(runtime));

	debug("%s request: OK", cmd);

	io_loop_run(io);
}

/* Connect to hal image manager. */
void hal_runtime_connect()
{
	hal_runtime_cmd(HAL_CMD_CONNECT);
}

/* Disconnect from hal image manager. */
void hal_runtime_disconnect()
{
	hal_runtime_cmd(HAL_CMD_DISCONNECT);
}

/* Send cycle begin request command. Then we wait input and output images. */
void hal_runtime_cycle_begin()
{
	hal_runtime_cmd(HAL_CMD_CYCLE_BEGIN);
}

/* Send cycle run request command. After calling this function we run our
 * runtime programm. */
void hal_runtime_cycle_run()
{
	hal_runtime_cmd(HAL_CMD_CYCLE_RUN);
}

/* Send cycle end request command. Then we write updated inputs and outputs
 * images back to the hal image manager. */
void hal_runtime_cycle_end()
{
	hal_runtime_cmd(HAL_CMD_CYCLE_END);
}
/* Below you can see reply handlers for hal commands parser. */
static void connect_reply_cb(unsigned int inputs_sz, unsigned int outputs_sz,
			     char side, const char *error, void *opaque)
{
	struct hal_runtime *runtime;

	runtime = (struct hal_runtime *)opaque;

	debug("CN reply");

	if (timeout_active(runtime->tout))
		timeout_del(runtime->tout);

	if (runtime->state != HAL_RT_STATE_CONN_WAIT) {
		runtime->state = HAL_RT_STATE_ERROR;
		warn("CN reply: runtime unexpected state");
		return;
	}

	if (error != NULL) {
		runtime->state = HAL_RT_STATE_ERROR;
		warn("CN reply: %s", error);
		return;
	}

	assert(runtime->callbacks.connect_callback != NULL);

	runtime->callbacks.connect_callback(inputs_sz, outputs_sz);
	runtime->state = HAL_RT_STATE_CONN;

	debug("CN reply: inputs size %d outputs size %d", inputs_sz, outputs_sz);
}

static void common_reply(const char *cmd, const char *error, void *opaque)
{
	struct hal_runtime *runtime;
	void (*user_cb)(void);
	int state;

	runtime = (struct hal_runtime *)opaque;

	debug("%s reply", cmd);

	if (timeout_active(runtime->tout))
		timeout_del(runtime->tout);

	user_cb = NULL;

	if (cmd == HAL_CMD_CYCLE_BEGIN) {
		state = HAL_RT_STATE_CB_WAIT;
		user_cb = runtime->callbacks.cycle_begin_callback;
	} else if (cmd == HAL_CMD_CYCLE_RUN) {
		state = HAL_RT_STATE_CR_WAIT;
		user_cb = runtime->callbacks.cycle_run_callback;
	} else if (cmd == HAL_CMD_CYCLE_END) {
		state = HAL_RT_STATE_CE_WAIT;
		user_cb = runtime->callbacks.cycle_end_callback;
	} else if (cmd == HAL_CMD_DISCONNECT) {
		state = HAL_RT_STATE_DISCONN_WAIT;
		user_cb = runtime->callbacks.disconnect_callback;
	}

	if (runtime->state != state) {
		runtime->state = HAL_RT_STATE_ERROR;
		warn("%s reply: runtime unexpected state", cmd);
		return;
	}

	if (error != NULL) {
		runtime->state = HAL_RT_STATE_ERROR;
		warn("%s reply: %s", cmd, error);
		return;
	}
	
	assert(user_cb != NULL);
	user_cb();
	runtime->state = HAL_RT_STATE_CONN;
}

static void cycle_begin_reply_cb(const char *error, void *opaque)
{
	common_reply(HAL_CMD_CYCLE_BEGIN, error, opaque);
}

static void cycle_run_reply_cb(const char *error, void *opaque)
{
	common_reply(HAL_CMD_CYCLE_RUN, error, opaque);
}

static void cycle_end_reply_cb(const char *error, void *opaque)
{
	common_reply(HAL_CMD_CYCLE_END, error, opaque);
}

static void disconnect_reply_cb(const char *error, void *opaque)
{
	common_reply(HAL_CMD_DISCONNECT, error, opaque);
}

/* */
static void write_image_request(const char *cmd, const char image[HAL_IMAGE_MAX+1],
				unsigned int off, void *opaque)
{
	const char *(*user_cb)(const char *, unsigned int);
	void (*write_reply)(struct buffer *);
	struct hal_runtime *runtime;
	const char *status;

	runtime = (struct hal_runtime *)opaque;

	debug("%s request", cmd);

	if (runtime->state != HAL_RT_STATE_CB_WAIT) {
		runtime->state = HAL_RT_STATE_ERROR;
		hal_status_error_reply(HAL_PEER_BASE(runtime)->wb, cmd,
					"Runtime unexpected state");
		goto send;
	}

	user_cb = NULL;

	if (cmd == HAL_CMD_WRITE_INPUTS) {
		user_cb = runtime->callbacks.write_inputs_callback;
		write_reply = hal_cmd_write_inputs_reply;
	} else if (cmd == HAL_CMD_WRITE_OUTPUTS) {
		user_cb = runtime->callbacks.write_outputs_callback;
		write_reply = hal_cmd_write_outputs_reply;
	}

	assert(user_cb != NULL);

	status = user_cb(image, off);
	if (status == HAL_OK)
		write_reply(HAL_PEER_BASE(runtime)->wb);
	else
		hal_status_error_reply(HAL_PEER_BASE(runtime)->wb, cmd, status);

	debug("%s request: status %s", cmd, status);
send:
	hal_peer_send_default(HAL_PEER_BASE(runtime));
}

static void write_inputs_request_cb(const char inputs[HAL_IMAGE_MAX+1],
				    unsigned int off, void *opaque)
{
	write_image_request(HAL_CMD_WRITE_INPUTS, inputs, off, opaque);
}

static void write_outputs_request_cb(const char outputs[HAL_IMAGE_MAX+1],
				     unsigned int off, void *opaque)
{
	write_image_request(HAL_CMD_WRITE_OUTPUTS, outputs, off, opaque);
}

static void read_image_request(const char *cmd, unsigned int len,
			       unsigned int off, void *opaque)
{
	const char *(*user_cb)(char *, unsigned int, unsigned int);
	void (*read_reply)(struct buffer *, unsigned int , const char *);
	struct hal_runtime *runtime;
	char image[HAL_IMAGE_MAX+1];
	const char *status;

	runtime = (struct hal_runtime *)opaque;

	debug("%s request", cmd);

	if (runtime->state != HAL_RT_STATE_CE_WAIT) {
		runtime->state = HAL_RT_STATE_ERROR;
		hal_status_error_reply(HAL_PEER_BASE(runtime)->wb, cmd,
					"Runtime unexpected state");
		goto send;
	}

	user_cb = NULL;
	/* Get the right read callback. */
	if (cmd == HAL_CMD_READ_INPUTS) {
		user_cb = runtime->callbacks.read_inputs_callback;
		read_reply = hal_cmd_read_inputs_reply;
	} else if (cmd == HAL_CMD_READ_OUTPUTS) {
		user_cb = runtime->callbacks.read_outputs_callback;
		read_reply = hal_cmd_read_outputs_reply;
	}

	assert(user_cb != NULL);

	status = user_cb(image, len, off);
	if (status == HAL_OK)
		read_reply(HAL_PEER_BASE(runtime)->wb, off, image);
	else
		hal_status_error_reply(HAL_PEER_BASE(runtime)->wb, cmd, status);
	
	debug("%s request: status %s", cmd, status);
send:
	hal_peer_send_default(HAL_PEER_BASE(runtime));
}

static void read_inputs_request_cb(unsigned int len, unsigned int off,
				   void *opaque)
{
	read_image_request(HAL_CMD_READ_INPUTS, len, off, opaque);
}

static void read_outputs_request_cb(unsigned int len, unsigned int off,
				    void *opaque)
{
	read_image_request(HAL_CMD_READ_OUTPUTS, len, off, opaque);
}

static void not_handled_cb()
{
	warn("Runtime get unexpected command");
}

static void parser_init()
{
	struct hal_cmd_callbacks parser_callbacks;

	parser_callbacks.connect_reply_callback = connect_reply_cb;
	parser_callbacks.disconnect_reply_callback = disconnect_reply_cb;
	parser_callbacks.cycle_begin_reply_callback = cycle_begin_reply_cb; 
	parser_callbacks.cycle_run_reply_callback = cycle_run_reply_cb; 
	parser_callbacks.cycle_end_reply_callback = cycle_end_reply_cb; 
	parser_callbacks.write_inputs_request_callback = write_inputs_request_cb;
	parser_callbacks.write_outputs_request_callback = write_outputs_request_cb;
	parser_callbacks.read_inputs_request_callback = read_inputs_request_cb;
	parser_callbacks.read_outputs_request_callback = read_outputs_request_cb;
	parser_callbacks.not_handled_callback = not_handled_cb;

	hal_cmd_parser_register_callbacks(&parser_callbacks);
}

static void peer_event_handler(struct hal_peer_base *peer,
			       hal_peer_event_t event)
{       
	struct hal_runtime *runtime;

	runtime = (struct hal_runtime *)peer;

	switch (event) {
	case HAL_PEER_EVENT_TIMEOUT:
		warn("got timeout event");
		runtime->state = HAL_RT_STATE_ERROR;
		break;
	case HAL_PEER_EVENT_ERROR:
		warn("got error event");
		runtime->state = HAL_RT_STATE_ERROR;
		break;
	default:
		break;
	}

	switch (runtime->state) {
	case HAL_RT_STATE_CONN:
	case HAL_RT_STATE_DISCONN:
	case HAL_RT_STATE_ERROR:
	case HAL_RT_STATE_DISABLED:
		io_loop_stop(peer->io);
		break;
	}
}

int hal_runtime_init(struct io_context *io_ptr, int wfd, int rfd, int timeout)
{
	struct hal_runtime *runtime;

	return_val_if_fail(io_ptr != NULL, -1);
	return_val_if_fail(io == NULL, -1);
	return_val_if_fail(hal_runtime == NULL, -1);
	return_val_if_fail(wfd >= 0, -1);
	return_val_if_fail(rfd >= 0, -1);

	runtime = xmalloc0(sizeof(*runtime));
	hal_peer_base_init(HAL_PEER_BASE(runtime), io_ptr, wfd, rfd, peer_event_handler);

	if (timeout > 0)
		runtime->response_timeout = timeout;
	else
		runtime->response_timeout = PEER_TIMEOUT_MSEC;

	runtime->tout = timeout_new(hal_peer_event_default, runtime);
	runtime->state = HAL_RT_STATE_DISCONN;

	hal_runtime = runtime;
	io = io_ptr;
	parser_init();

	return 0;
}

int hal_runtime_register_callbacks(struct hal_runtime_callbacks *new_cb)
{	
	return_val_if_fail(new_cb != NULL, -1);
	return_val_if_fail(hal_runtime != NULL, -1);

	hal_runtime->callbacks = *new_cb;

	return 0;
}

void hal_runtime_deinit()
{
	return_if_fail(io != NULL);
	
	if (timeout_active(hal_runtime->tout))
		timeout_del(hal_runtime->tout);

	timeout_free(&hal_runtime->tout);

	hal_peer_base_deinit(HAL_PEER_BASE(hal_runtime));

	xfree(hal_runtime);
	io = NULL;
	hal_runtime = NULL;
}

